# -*- coding: utf-8 -*-
#
# Copyright © 2018 André Roberge - mod_pydoc
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""PyDoc patch"""

# Standard libray
import builtins
import io
import inspect
import os
import pkgutil
import platform
from pydoc import (
    classname, classify_class_attrs, describe, Doc, format_exception_only,
    Helper, HTMLRepr, _is_bound_method, ModuleScanner, locate, replace,
    visiblename, isdata, getdoc, deque, _split_list)
import re
import sys
import tokenize
import warnings

# Local imports
from spyder.api.fonts import SpyderFontsMixin, SpyderFontType
from spyder.api.translations import _
from spyder.config.base import DEV
from spyder.config.gui import is_dark_interface


class CustomHTMLDoc(Doc, SpyderFontsMixin):
    """
    Formatter class for HTML documentation.

    See:
    https://github.com/aroberge/mod_pydoc/blob/master/mod_pydoc/pydoc.py
    """

    # ------------------------------------------ HTML formatting utilities

    _repr_instance = HTMLRepr()
    repr = _repr_instance.repr
    escape = _repr_instance.escape

    def __init__(self):
        super().__init__()
        self.app_font = self.get_font(SpyderFontType.Interface)
        self.plain_font = self.get_font(SpyderFontType.Monospace)

        # The increased value used below is necessary to make the font size on
        # this plugin match the one used in the rest of the app.
        self.app_font_size = self.app_font.pointSize() + 2

    def page(self, title, contents):
        """Format an HTML page."""
        return '''\
<!doctype html>
<html><head><title>Python: %s</title>
<meta charset="UTF-8">
</head><body>
%s
</body></html>''' % (title, contents)

    def heading(self, title, extras=''):
        """Format a page heading."""
        return '''
<table class="heading">
<tr><td>{}</td><td class="align_right normal">{}</td></tr></table>
    '''.format(title, extras or '&nbsp;')

    def html_section(self, title, contents, width=6,
                     prelude='', marginalia=None, gap='&nbsp;',
                     css_class=''):
        """Format a section with a heading."""
        result = '''
<table class="{}" style="font-size: {}pt">
<tr>
<td colspan="3">
{}</td></tr>
    '''.format(css_class, self.app_font_size, title)

        if prelude:
            result = result + '''
<tr><td rowspan="2">{}</td>
<td colspan="2">{}</td></tr>
<tr><td>{}</td>'''.format(marginalia, prelude, gap)
        elif marginalia:
            result = result + '''
<tr><td>{}</td><td>{}</td>'''.format(marginalia, gap)

        contents = '{}</td></tr></table><br>'.format(contents)
        return result + '\n<td class="inner_table">' + contents

    def bigsection(self, title, *args, **kwargs):
        """Format a section with a big heading."""
        title = '<span class="section_title">{}</span>'.format(title)
        return self.html_section(title, *args, **kwargs)

    def preformat(self, text):
        """Format literal preformatted text."""
        text = self.escape(text.expandtabs())
        return replace(text, '\n\n', '\n \n', '\n\n', '\n \n',
                             ' ', '&nbsp;', '\n', '<br>\n')

    def multicolumn(self, list, format, cols=4):
        """Format a list of items into a multi-column list."""
        result = ''
        rows = (len(list)+cols-1)//cols
        for col in range(cols):
            result = (
                result + '<td style="width:%d%%;vertical-align:text-top">'
                % (100//cols))
            for i in range(rows*col, rows*col+rows):
                if i < len(list):
                    result = result + format(list[i]) + '<br>\n'
            result = result + '</td>'
        return (
            '<table style="width:100%; font-size: {}pt">'
            '<tr>{}</tr></table>'
        ).format(self.app_font_size, result)

    def grey(self, text):
        """Grey span."""
        return '<span class="grey">%s</span>' % text

    def namelink(self, name, *dicts):
        """Make a link for an identifier, given name-to-URL mappings."""
        for dict in dicts:
            if name in dict:
                return '<a href="%s">%s</a>' % (dict[name], name)
        return name

    def classlink(self, object, modname):
        """Make a link for a class."""
        name, module = object.__name__, sys.modules.get(object.__module__)
        if hasattr(module, name) and getattr(module, name) is object:
            return '<a href="%s.html#%s">%s</a>' % (
                module.__name__, name, classname(object, modname))
        return classname(object, modname)

    def modulelink(self, object):
        """Make a link for a module."""
        return '<a href="%s.html">%s</a>' % (
            object.__name__, object.__name__)

    def modpkglink(self, modpkginfo):
        """Make a link for a module or package to display in an index."""
        name, path, ispackage, shadowed = modpkginfo
        if shadowed:
            return self.grey(name)
        if path:
            url = '%s.%s.html' % (path, name)
        else:
            url = '%s.html' % name
        if ispackage:
            text = '%s&nbsp;(package)' % name
        else:
            text = name
        return '<a href="%s">%s</a>' % (url, text)

    def filelink(self, url, path):
        """Make a link to source file."""
        return '<a href="file:%s">%s</a>' % (url, path)

    def markup(
        self, text, escape=None, funcs=None, classes=None, methods=None
    ):
        """
        Mark up some plain text, given a context of symbols to look for.

        Each context dictionary maps object names to anchor names.
        """
        funcs = {} if funcs is None else funcs
        classes = {} if classes is None else classes
        methods = {} if methods is None else methods
        escape = escape or self.escape
        results = []
        here = 0
        pattern = re.compile(r'\b((http|ftp)://\S+[\w/]|'
                             r'RFC[- ]?(\d+)|'
                             r'PEP[- ]?(\d+)|'
                             r'(self\.)?(\w+))')
        while True:
            match = pattern.search(text, here)
            if not match:
                break
            start, end = match.span()
            results.append(escape(text[here:start]))

            all, scheme, rfc, pep, selfdot, name = match.groups()
            if scheme:
                url = escape(all).replace('"', '&quot;')
                results.append('<a href="%s">%s</a>' % (url, url))
            elif rfc:
                url = 'http://www.rfc-editor.org/rfc/rfc%d.txt' % int(rfc)
                results.append('<a href="%s">%s</a>' % (url, escape(all)))
            elif pep:
                url = 'http://www.python.org/dev/peps/pep-%04d/' % int(pep)
                results.append('<a href="%s">%s</a>' % (url, escape(all)))
            elif text[end:end+1] == '(':
                results.append(
                    self.namelink(name, methods, funcs, classes))
            elif selfdot:
                results.append('self.%s' % name)
            else:
                results.append(self.namelink(name, classes))
            here = end

        results.append(escape(text[here:]))
        return ''.join(results)

    # --------------------------------------------- type-specific routines

    def formattree(self, tree, modname, parent=None):
        """
        Produce HTML for a class tree as given by inspect.getclasstree().
        """
        result = ''
        for entry in tree:
            if type(entry) is type(()):
                c, bases = entry
                result = result + '<dt>'
                result = result + self.classlink(c, modname)
                if bases and bases != (parent,):
                    parents = []
                    for base in bases:
                        parents.append(self.classlink(base, modname))
                    result = result + '(' + ', '.join(parents) + ')'
                result = result + '\n</dt>'
            elif type(entry) is type([]):
                result = result + '<dd>\n%s</dd>\n' % self.formattree(
                    entry, modname, c)
        return '<dl><dt></dt>\n%s<dd></dd></dl>\n' % result

    def docmodule(self, object, name=None, mod=None, *ignored):
        """Produce HTML documentation for a module object."""
        name = object.__name__  # ignore the passed-in name

        try:
            all = object.__all__
        except AttributeError:
            all = None

        parts = name.split('.')
        links = []
        for i in range(len(parts)-1):
            links.append(
                '<a href="{}.html" class="docmodule_link">{}</a>'.format(
                     '.'.join(parts[:i+1]), parts[i]))
        head = '.'.join(links + parts[-1:])

        try:
            path = inspect.getabsfile(object)
            url = path
            if sys.platform == 'win32':
                import nturl2path
                url = nturl2path.pathname2url(path)
            filelink = self.filelink(url, path)
        except TypeError:
            filelink = '(built-in)'

        info = []
        if hasattr(object, '__version__'):
            version = str(object.__version__)
            if version[:11] == '$' + 'Revision: ' and version[-1:] == '$':
                version = version[11:-1].strip()
            info.append('version %s' % self.escape(version))
        if hasattr(object, '__date__'):
            info.append(self.escape(str(object.__date__)))
        if info:
            head = head + ' (%s)' % ', '.join(info)

        docloc = self.getdocloc(object)
        if docloc is not None:
            docloc = (
                '<br><a href="%(docloc)s">Module Reference</a>' % locals())
        else:
            docloc = ''
        extras = '<a href=".">index</a><br>' + filelink + docloc
        result = self.heading(head, extras)

        modules = inspect.getmembers(object, inspect.ismodule)

        classes, cdict = [], {}
        for key, value in inspect.getmembers(object, inspect.isclass):
            # if __all__ exists, believe it.  Otherwise use old heuristic.
            if (all is not None or
                    (inspect.getmodule(value) or object) is object):
                if visiblename(key, all, object):
                    classes.append((key, value))
                    cdict[key] = cdict[value] = '#' + key

        for key, value in classes:
            for base in value.__bases__:
                key, modname = base.__name__, base.__module__
                module = sys.modules.get(modname)
                if modname != name and module and hasattr(module, key):
                    if getattr(module, key) is base:
                        if key not in cdict:
                            cdict[key] = cdict[base] = (
                                modname + '.html#' + key)

        funcs, fdict = [], {}
        for key, value in inspect.getmembers(object, inspect.isroutine):
            # if __all__ exists, believe it.  Otherwise use old heuristic.
            if (all is not None or
                    inspect.isbuiltin(value) or
                    inspect.getmodule(value) is object):
                if visiblename(key, all, object):
                    funcs.append((key, value))
                    fdict[key] = '#-' + key
                    if inspect.isfunction(value):
                        fdict[value] = fdict[key]

        data = []
        for key, value in inspect.getmembers(object, isdata):
            if visiblename(key, all, object):
                data.append((key, value))

        doc = self.markup(getdoc(object), self.preformat, fdict, cdict)
        doc = doc and '<code>{}</code>'.format(doc)
        result = result + '<p>%s</p>\n' % doc

        if hasattr(object, '__path__'):
            modpkgs = []
            for importer, modname, ispkg in pkgutil.iter_modules(
                    object.__path__):
                modpkgs.append((modname, name, ispkg, 0))
            modpkgs.sort()
            contents = self.multicolumn(modpkgs, self.modpkglink)
            result = result + self.bigsection(
                'Package Contents', contents, css_class="package")
        elif modules:
            contents = self.multicolumn(
                modules, lambda t: self.modulelink(t[1]))
            result = result + self.bigsection(
                'Modules', contents, css_class="module")

        if classes:
            classlist = [value for (key, value) in classes]
            contents = [
                self.formattree(inspect.getclasstree(classlist, 1), name)]
            for key, value in classes:
                contents.append(
                    self.document(value, key, name, fdict, cdict))
            result = result + self.bigsection(
                'Classes', ' '.join(contents), css_class="classes")

        if funcs:
            contents = []
            for key, value in funcs:
                contents.append(
                    self.document(value, key, name, fdict, cdict))
            result = result + self.bigsection(
                'Functions', ' '.join(contents), css_class="functions")

        if data:
            contents = []
            for key, value in data:
                contents.append(self.document(value, key))
            result = result + self.bigsection(
                'Data', '<br>\n'.join(contents), css_class="data")

        if hasattr(object, '__author__'):
            contents = self.markup(str(object.__author__), self.preformat)
            result = result + self.bigsection(
                'Author', contents, css_class="author")

        if hasattr(object, '__credits__'):
            contents = self.markup(str(object.__credits__), self.preformat)
            result = result + self.bigsection(
                'Credits', contents, css_class="credits")

        return result

    def docclass(self, object, name=None, mod=None, funcs=None, classes=None,
                 *ignored):
        """Produce HTML documentation for a class object."""
        funcs = {} if funcs is None else funcs
        classes = {} if classes is None else classes
        realname = object.__name__
        name = name or realname
        bases = object.__bases__

        contents = []
        push = contents.append

        # Cute little class to pump out a horizontal rule between sections.
        class HorizontalRule:
            def __init__(self):
                self.needone = 0

            def maybe(self):
                if self.needone:
                    push('<hr>\n')
                self.needone = 1
        hr = HorizontalRule()

        # List the mro, if non-trivial.
        mro = deque(inspect.getmro(object))
        if len(mro) > 2:
            hr.maybe()
            push('<dl><dt>Method resolution order:</dt>\n')
            for base in mro:
                push('<dd>%s</dd>\n' % self.classlink(base,
                                                      object.__module__))
            push('</dl>\n')

        def spill(msg, attrs, predicate):
            ok, attrs = _split_list(attrs, predicate)
            if ok:
                hr.maybe()
                push(msg)
                for name, kind, homecls, value in ok:
                    try:
                        value = getattr(object, name)
                    except Exception:
                        # Some descriptors may meet a failure
                        # in their __get__.
                        # (bug aroberge/mod_pydoc#1785)
                        push(self._docdescriptor(name, value, mod))
                    else:
                        push(self.document(
                            value, name, mod, funcs, classes, mdict,
                            object))
                    push('\n')
            return attrs

        def spilldescriptors(msg, attrs, predicate):
            ok, attrs = _split_list(attrs, predicate)
            if ok:
                hr.maybe()
                push(msg)
                for name, kind, homecls, value in ok:
                    push(self._docdescriptor(name, value, mod))
            return attrs

        def spilldata(msg, attrs, predicate):
            ok, attrs = _split_list(attrs, predicate)
            if ok:
                hr.maybe()
                push(msg)
                for name, kind, homecls, value in ok:
                    base = self.docother(getattr(object, name), name, mod)
                    if callable(value) or inspect.isdatadescriptor(value):
                        doc = getattr(value, "__doc__", None)
                    else:
                        doc = None
                    if doc is None:
                        push('<dl><dt>%s</dt><dd></dd></dl>\n' % base)
                    else:
                        doc = self.markup(getdoc(value), self.preformat,
                                          funcs, classes, mdict)
                        doc = '<dd><code>%s</code></dd>' % doc
                        push('<dl><dt>%s%s</dt></dl>\n' % (base, doc))
                    push('\n')
            return attrs

        attrs = [(name, kind, cls, value)
                 for name, kind, cls, value in classify_class_attrs(object)
                 if visiblename(name, obj=object)]

        mdict = {}
        for key, kind, homecls, value in attrs:
            mdict[key] = anchor = '#' + name + '-' + key
            try:
                value = getattr(object, name)
            except Exception:
                # Some descriptors may meet a failure in their __get__.
                # (bug #1785)
                pass
            try:
                # The value may not be hashable (e.g., a data attr with
                # a dict or list value).
                mdict[value] = anchor
            except TypeError:
                pass

        while attrs:
            if mro:
                thisclass = mro.popleft()
            else:
                thisclass = attrs[0][2]
            attrs, inherited = _split_list(
                attrs, lambda t: t[2] is thisclass)

            if thisclass is builtins.object:
                attrs = inherited
                continue
            elif thisclass is object:
                tag = 'defined here'
            else:
                tag = 'inherited from %s' % self.classlink(
                    thisclass, object.__module__)
            tag += ':<br>\n'

            # Sort attrs by name.
            attrs.sort(key=lambda t: t[0])

            # Pump out the attrs, segregated by kind.
            attrs = spill('Methods %s' % tag, attrs,
                          lambda t: t[1] == 'method')
            attrs = spill('Class methods %s' % tag, attrs,
                          lambda t: t[1] == 'class method')
            attrs = spill('Static methods %s' % tag, attrs,
                          lambda t: t[1] == 'static method')
            attrs = spilldescriptors("Readonly properties %s" % tag, attrs,
                                     lambda t: t[1] == 'readonly property')
            attrs = spilldescriptors('Data descriptors %s' % tag, attrs,
                                     lambda t: t[1] == 'data descriptor')
            attrs = spilldata('Data and other attributes %s' % tag, attrs,
                              lambda t: t[1] == 'data')
            assert attrs == []
            attrs = inherited

        contents = ''.join(contents)

        if name == realname:
            title = '<span id="%s" class="signature"> class %s</span>' % (
                name, realname)
        else:
            title = (
                '%s = <span id="%s" class="signature">class %s</span>' % (
                    name, name, realname))
        if bases:
            parents = []
            for base in bases:
                parents.append(self.classlink(base, object.__module__))
            title = title + '(%s)' % ', '.join(parents)
        doc = self.markup(
            getdoc(object), self.preformat, funcs, classes, mdict)
        doc = doc and '<code>%s<br>&nbsp;</code>' % doc

        return self.html_section(
            title, contents, 3, doc, css_class="docclass")

    def formatvalue(self, object):
        """Format an argument default value as text."""
        return self.grey('=' + self.repr(object))

    def docroutine(self, object, name=None, mod=None,
                   funcs={}, classes={}, methods={}, cl=None):
        """Produce HTML documentation for a function or method object."""
        realname = object.__name__
        name = name or realname
        anchor = (cl and cl.__name__ or '') + '-' + name
        note = ''
        skipdocs = 0
        if _is_bound_method(object):
            imclass = object.__self__.__class__
            if cl:
                if imclass is not cl:
                    note = ' from ' + self.classlink(imclass, mod)
            else:
                if object.__self__ is not None:
                    note = ' method of %s instance' % self.classlink(
                        object.__self__.__class__, mod)
                else:
                    note = ' unbound %s method' % self.classlink(
                        imclass, mod)

        if name == realname:
            title = '<span id="%s" class="signature">%s</span>' % (
                anchor, realname)
        else:
            if (cl and realname in cl.__dict__ and
                    cl.__dict__[realname] is object):
                reallink = '<a href="#%s">%s</a>' % (
                    cl.__name__ + '-' + realname, realname)
                skipdocs = 1
            else:
                reallink = realname
            title = '<span id="%s" class="signature">%s</span> = %s' % (
                anchor, name, reallink)
        argspec = None
        if inspect.isroutine(object):
            try:
                signature = inspect.signature(object)
            except (ValueError, TypeError):
                signature = None
            if signature:
                argspec = str(signature)
                if realname == '<lambda>':
                    title = '%s <em>lambda</em> ' % name
                    # XXX lambda's won't usually have
                    # func_annotations['return']
                    # since the syntax doesn't support but it is possible.
                    # So removing parentheses isn't truly safe.
                    argspec = argspec[1:-1]  # remove parentheses
        if not argspec:
            argspec = '(...)'

        decl = title + argspec + (note and self.grey(note))

        if skipdocs:
            return '<dl><dt>%s</dt><dd></dd></dl>\n' % decl
        else:
            doc = self.markup(
                getdoc(object), self.preformat, funcs, classes, methods)
            doc = doc and '<dd><code>%s</code></dd>' % doc
            return '<dl><dt>%s</dt><dd></dd>%s</dl>\n' % (decl, doc)

    def _docdescriptor(self, name, value, mod):
        results = []
        push = results.append

        if name:
            push('<dl><dt>%s</dt>\n' % name)
        if value.__doc__ is not None:
            doc = self.markup(getdoc(value), self.preformat)
            push('<dd><code>%s</code></dd>\n' % doc)
        push('<dd></dd></dl>\n')

        return ''.join(results)

    def docproperty(self, object, name=None, mod=None, cl=None):
        """Produce html documentation for a property."""
        return self._docdescriptor(name, object, mod)

    def docother(self, object, name=None, mod=None, *ignored):
        """Produce HTML documentation for a data object."""
        lhs = name and '%s = ' % name or ''
        return lhs + self.repr(object)

    def docdata(self, object, name=None, mod=None, cl=None):
        """Produce html documentation for a data descriptor."""
        return self._docdescriptor(name, object, mod)

    def index(self, dir, shadowed=None):
        """Generate an HTML index for a directory of modules."""
        modpkgs = []
        if shadowed is None:
            shadowed = {}
        for importer, name, ispkg in pkgutil.iter_modules([dir]):
            if any((0xD800 <= ord(ch) <= 0xDFFF) for ch in name):
                # ignore a module if its name contains a
                # surrogate character
                continue
            modpkgs.append((name, '', ispkg, name in shadowed))
            shadowed[name] = 1

        modpkgs.sort()
        if len(modpkgs):
            contents = self.multicolumn(modpkgs, self.modpkglink)
            return self.bigsection(dir, contents, css_class="index")
        else:
            return ''


def _url_handler(url, content_type="text/html"):
    """Pydoc url handler for use with the pydoc server.

    If the content_type is 'text/css', the _pydoc.css style
    sheet is read and returned if it exits.

    If the content_type is 'text/html', then the result of
    get_html_page(url) is returned.

    See https://github.com/python/cpython/blob/master/Lib/pydoc.py
    """
    class _HTMLDoc(CustomHTMLDoc):

        def page(self, title, contents):
            """Format an HTML page."""

            if is_dark_interface():
                css_path = "static/css/dark_pydoc.css"
            else:
                css_path = "static/css/light_pydoc.css"

            css_link = (
                f'<link rel="stylesheet" type="text/css" href="/{css_path}">'
            )

            code_style = (
                '<style>code {{font-family: "{}"; font-size: {}pt}}</style>'
            ).format(self.plain_font.family(), self.app_font_size)

            html_page = '''\
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">
<html><head><title>Pydoc: {}</title>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8">
{}{}</head><body style="clear: both; font-family: '{}'; font-size: {}pt">
{}<div style="clear:both;padding-top:.7em;">{}</div>
</body></html>'''.format(
    title, css_link, code_style, self.app_font.family(),
    self.app_font_size, html_navbar(), contents)

            return html_page

        def filelink(self, url, path):
            return '<a href="getfile?key=%s">%s</a>' % (url, path)

    html = _HTMLDoc()

    def html_navbar():
        version = html.escape("%s [%s, %s]" % (platform.python_version(),
                                               platform.python_build()[0],
                                               platform.python_compiler()))
        return """
            <div style='float:left'>
                Python %s<br>%s
            </div>
            <div style='float:right'>
                <div style='text-align:right; padding-bottom:.7em;'>
                  <a href="index.html">Module Index</a>
                  : <a href="topics.html">Topics</a>
                  : <a href="keywords.html">Keywords</a>
                </div>
                <div style='text-align:right;'>
                    <form action="search" style='display:inline;'>
                      <input class="input-search" type=text name=key size="22">
                      <input class="submit-search" type=submit value="Search">
                    </form>
                </div>
            </div>
            """ % (version, html.escape(platform.platform(terse=True)))

    def html_index():
        """Index page."""
        def bltinlink(name):
            return '<a href="%s.html">%s</a>' % (name, name)

        heading = html.heading('<span>Index of Modules</span>')
        names = [name for name in sys.builtin_module_names
                 if name != '__main__']
        contents = html.multicolumn(names, bltinlink)
        contents = [heading, '<p>' + html.bigsection(
            'Built-in Modules', contents, css_class="builtin_modules")]

        seen = {}
        for dir in sys.path:
            contents.append(html.index(dir, seen))

        return 'Index of Modules', ''.join(contents)

    def html_search(key):
        """Search results page."""
        # scan for modules
        search_result = []

        def callback(path, modname, desc):
            if modname[-9:] == '.__init__':
                modname = modname[:-9] + ' (package)'
            search_result.append((modname, desc and '- ' + desc))

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')  # ignore problems during import
            def onerror(modname):
                pass
            ModuleScanner().run(callback, key, onerror=onerror)

        # format page
        def bltinlink(name):
            return '<a href="%s.html">%s</a>' % (name, name)

        results = []
        heading = html.heading('Search Results')

        for name, desc in search_result:
            results.append(bltinlink(name) + desc)
        contents = heading + html.bigsection(
            'key = {}'.format(key), '<br>'.join(results), css_class="search")
        return 'Search Results', contents

    def html_getfile(path):
        """Get and display a source file listing safely."""
        path = path.replace('%20', ' ')
        with tokenize.open(path) as fp:
            lines = html.escape(fp.read())
        body = '<pre>%s</pre>' % lines
        heading = html.heading('File Listing')

        contents = heading + html.bigsection('File: {}'.format(path), body,
                                             css_class="getfile")
        return 'getfile %s' % path, contents

    def html_topics():
        """Index of topic texts available."""
        def bltinlink(name):
            return '<a href="topic?key=%s">%s</a>' % (name, name)

        heading = html.heading('Index of Topics') + '<br>'
        names = sorted(Helper.topics.keys())

        contents = html.multicolumn(names, bltinlink)
        contents = heading + html.bigsection(
            'Topics', contents, css_class="topics")
        return 'Topics', contents

    def html_keywords():
        """Index of keywords."""
        heading = html.heading('Index of Keywords')
        names = sorted(Helper.keywords.keys())

        def bltinlink(name):
            return '<a href="topic?key=%s">%s</a>' % (name, name)

        contents = html.multicolumn(names, bltinlink)
        contents = heading + '<br>' + html.bigsection(
            'Keywords', contents, css_class="keywords")
        return 'Keywords', contents

    def html_topicpage(topic):
        """Topic or keyword help page."""
        buf = io.StringIO()
        htmlhelp = Helper(buf, buf)
        contents, xrefs = htmlhelp._gettopic(topic)
        if topic in htmlhelp.keywords:
            title = 'Keyword'
        else:
            title = 'Topic'
        heading = html.heading(title)
        contents = '<pre>%s</pre>' % html.markup(contents)
        contents = html.bigsection(topic, contents, css_class="topics")
        if xrefs:
            xrefs = sorted(xrefs.split())

            def bltinlink(name):
                return '<a href="topic?key=%s">%s</a>' % (name, name)

            xrefs = html.multicolumn(xrefs, bltinlink)
            xrefs = html.html_section('Related help topics: ', xrefs,
                                      css_class="topics")
        return ('%s %s' % (title, topic),
                ''.join((heading, contents, xrefs)))

    def html_getobj(url):
        obj = locate(url, forceload=1)
        if obj is None and url != 'None':
            raise ValueError(
                _('There was an error while retrieving documentation '
                  'for the object you requested: Object could not be found'))
        title = describe(obj)
        content = html.document(obj, url)
        return title, content

    def html_error(url, exc):
        heading = html.heading('Error')
        if DEV:
            contents = '<br>'.join(html.escape(line) for line in
                                   format_exception_only(type(exc), exc))
        else:
            contents = '%s' % str(exc)
        contents = heading + html.bigsection(url, contents, css_class="error")
        return "Error - %s" % url, contents

    def get_html_page(url):
        """Generate an HTML page for url."""
        complete_url = url
        if url.endswith('.html'):
            url = url[:-5]
        try:
            if url in ("", "index"):
                title, content = html_index()
            elif url == "topics":
                title, content = html_topics()
            elif url == "keywords":
                title, content = html_keywords()
            elif '=' in url:
                op, _, url = url.partition('=')
                if op == "search?key":
                    title, content = html_search(url)
                elif op == "getfile?key":
                    title, content = html_getfile(url)
                elif op == "topic?key":
                    # try topics first, then objects.
                    try:
                        title, content = html_topicpage(url)
                    except ValueError:
                        title, content = html_getobj(url)
                elif op == "get?key":
                    # try objects first, then topics.
                    if url in ("", "index"):
                        title, content = html_index()
                    else:
                        try:
                            title, content = html_getobj(url)
                        except ValueError:
                            title, content = html_topicpage(url)
                else:
                    raise ValueError(
                        _('There was an error while retrieving documentation '
                          'for the object you requested: Bad URL %s') % url)
            else:
                title, content = html_getobj(url)
        except Exception as exc:
            # Catch any errors and display them in an error page.
            title, content = html_error(complete_url, exc)
        return html.page(title, content)

    if url.startswith('/'):
        url = url[1:]
    if content_type == 'text/css':
        path_here = os.path.dirname(os.path.realpath(__file__))
        css_path = os.path.join(path_here, url)
        with open(css_path) as fp:
            return ''.join(fp.readlines())
    elif content_type == 'text/html':
        return get_html_page(url)
    # Errors outside the url handler are caught by the server.
    raise TypeError(
        _('There was an error while retrieving documentation '
          'for the object you requested: unknown content type %r for url %s')
          % (content_type, url))


def _start_server(urlhandler, hostname, port):
    """
    Start an HTTP server thread on a specific port.

    This is a reimplementation of `pydoc._start_server` to handle connection
    errors for 'do_GET'.

    Taken from PyDoc: https://github.com/python/cpython/blob/3.7/Lib/pydoc.py
    """
    import http.server
    import email.message
    import select
    import threading
    import time

    class DocHandler(http.server.BaseHTTPRequestHandler):

        def do_GET(self):
            """Process a request from an HTML browser.

            The URL received is in self.path.
            Get an HTML page from self.urlhandler and send it.
            """
            if self.path.endswith('.css'):
                content_type = 'text/css'
            else:
                content_type = 'text/html'
            self.send_response(200)
            self.send_header(
                'Content-Type', '%s; charset=UTF-8' % content_type)
            self.end_headers()
            try:
                self.wfile.write(self.urlhandler(
                    self.path, content_type).encode('utf-8'))
            except ConnectionAbortedError:
                # Needed to handle error when client closes the connection,
                # for example when the client stops the load of the previously
                # requested page. See spyder-ide/spyder#10755
                pass
            except BrokenPipeError:
                # Needed to handle permission error when trying to open a port
                # for the web server of the online help.
                # See spyder-ide/spyder#13388
                pass

        def log_message(self, *args):
            # Don't log messages.
            pass

    class DocServer(http.server.HTTPServer):

        def __init__(self, host, port, callback):
            self.host = host
            self.address = (self.host, port)
            self.callback = callback
            self.base.__init__(self, self.address, self.handler)
            self.quit = False

        def serve_until_quit(self):
            while not self.quit:
                rd, wr, ex = select.select([self.socket.fileno()], [], [], 1)
                if rd:
                    self.handle_request()
            self.server_close()

        def server_activate(self):
            self.base.server_activate(self)
            if self.callback:
                self.callback(self)

    class ServerThread(threading.Thread):

        def __init__(self, urlhandler, host, port):
            self.urlhandler = urlhandler
            self.host = host
            self.port = int(port)
            threading.Thread.__init__(self)
            self.serving = False
            self.error = None

        def run(self):
            """Start the server."""
            try:
                DocServer.base = http.server.HTTPServer
                DocServer.handler = DocHandler
                DocHandler.MessageClass = email.message.Message
                DocHandler.urlhandler = staticmethod(self.urlhandler)
                docsvr = DocServer(self.host, self.port, self.ready)
                self.docserver = docsvr
                docsvr.serve_until_quit()
            except Exception as e:
                self.error = e

        def ready(self, server):
            self.serving = True
            self.host = server.host
            self.port = server.server_port
            self.url = 'http://%s:%d/' % (self.host, self.port)

        def stop(self):
            """Stop the server and this thread nicely."""
            self.docserver.quit = True
            self.join()
            # explicitly break a reference cycle: DocServer.callback
            # has indirectly a reference to ServerThread.
            self.docserver = None
            self.serving = False
            self.url = None

    thread = ServerThread(urlhandler, hostname, port)
    thread.daemon = True
    thread.start()
    # Wait until thread.serving is True to make sure we are
    # really up before returning.
    while not thread.error and not thread.serving:
        time.sleep(.01)
    return thread



# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""Tests for the remote files API."""

# Third party imports
import pytest

from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.plugins.remoteclient.plugin import RemoteClient
from spyder.plugins.remoteclient.tests.conftest import mark_remote_test


@mark_remote_test
class TestRemoteEnvironAPI:
    test_key = "TEST_VAR"
    test_value = "test_value"

    @AsyncDispatcher(early_return=False)
    async def test_set_var(
        self,
        remote_client: RemoteClient,
        remote_client_id: str,
    ):
        """
        Test that an environment variable can be set on the remote server.
        """
        environ_api_class = remote_client.get_environ_api(remote_client_id)
        assert environ_api_class is not None

        async with environ_api_class() as environ_api:
            await environ_api.set(self.test_key, self.test_value)
            value = await environ_api.get(self.test_key)

        assert value == self.test_value

    @AsyncDispatcher(early_return=False)
    async def test_list_environ(
        self,
        remote_client: RemoteClient,
        remote_client_id: str,
    ):
        """
        Test that environment variables can be listed on the remote server.
        """
        environ_api_class = remote_client.get_environ_api(remote_client_id)
        assert environ_api_class is not None

        async with environ_api_class() as environ_api:
            env_vars = await environ_api.to_dict()
            assert isinstance(env_vars, dict)
            assert self.test_key in env_vars
            assert env_vars[self.test_key] == self.test_value

    @AsyncDispatcher(early_return=False)
    async def test_get_non_existent_var(
        self,
        remote_client: RemoteClient,
        remote_client_id: str,
    ):
        """
        Test that getting a non-existent environment variable returns None.
        """
        environ_api_class = remote_client.get_environ_api(remote_client_id)
        assert environ_api_class is not None

        async with environ_api_class() as environ_api:
            value = await environ_api.get("NON_EXISTENT_VAR")
            assert value is None

    @AsyncDispatcher(early_return=False)
    async def test_delete_var(
        self,
        remote_client: RemoteClient,
        remote_client_id: str,
    ):
        """
        Test that an environment variable can be deleted on the remote server.
        """
        environ_api_class = remote_client.get_environ_api(remote_client_id)
        assert environ_api_class is not None

        async with environ_api_class() as environ_api:
            await environ_api.delete(self.test_key)
            value = await environ_api.get(self.test_key)

        assert value is None


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""Fixtures for the Application plugin tests."""

# Standard library imports
from unittest.mock import Mock, patch

# Third party imports
import pytest

# Local imports
from spyder.config.base import running_in_ci
from spyder.config.manager import CONF
from spyder.plugins.application.plugin import Application
from spyder.utils.stylesheet import APP_STYLESHEET


@pytest.fixture
def application_plugin(qapp, qtbot):
    main_window = Mock()
    CONF.set('main', 'recent_files', [])
    application = Application(None, configuration=CONF)
    application.main = application._main = main_window
    application.on_initialize()
    container = application.get_container()
    container.setMinimumSize(700, 500)
    qtbot.addWidget(container)
    if not running_in_ci():
        qapp.setStyleSheet(str(APP_STYLESHEET))
    with qtbot.waitExposed(container):
        container.show()

    with patch.object(application, 'get_plugin'):
        yield application



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

# Standard library imports

# Third party imports
import pytest
from qtpy.QtCore import Qt, QPoint
from qtpy.QtGui import QTextCursor

# Local imports


ControlModifier = Qt.KeyboardModifier.ControlModifier
AltModifier = Qt.KeyboardModifier.AltModifier
ShiftModifier = Qt.KeyboardModifier.ShiftModifier


def click_at(codeeditor, qtbot, position, ctrl=False, alt=False, shift=False):
    """
    Convienience function to generate a mouseClick at a given text position
    with the specified modifiers held.
    """

    x, y = codeeditor.get_coordinates(position)
    point = QPoint(x, y)

    modifiers = Qt.KeyboardModifier.NoModifier
    if ctrl:
        modifiers |= ControlModifier
    if alt:
        modifiers |= AltModifier
    if shift:
        modifiers |= ShiftModifier

    qtbot.mouseClick(
        codeeditor.viewport(),
        Qt.MouseButton.LeftButton,
        modifiers,
        pos=point
    )


def call_shortcut(codeeditor, name):
    """
    Convienience function to call a QShortcut without having to simulate the
    key sequence (which may be different for each platform?)
    """
    context = codeeditor.CONF_SECTION.lower()
    plugin_name = None
    qshortcut = codeeditor._shortcuts[(context, name, plugin_name)]
    qshortcut.activated.emit()


@pytest.mark.order(1)
def test_add_cursor(codeeditor, qtbot):
    """Test adding and removing extra cursors with crtl-alt click"""

    # Enabled by default arg on CodeEditor.setup_editor (which is called in the
    # pytest fixture creation in conftest.py)
    assert codeeditor.multi_cursor_enabled  # This is assumed for other tests
    assert codeeditor.cursorWidth() == 0  # Required for multi-cursor rendering
    codeeditor.set_text("0123456789")
    click_at(codeeditor, qtbot, 6, ctrl=True, alt=True)

    # A cursor was added
    assert bool(codeeditor.extra_cursors)
    qtbot.keyClick(codeeditor, "a")

    # Text was inserted correctly from two cursors
    assert codeeditor.toPlainText() == "a012345a6789"

    # Regular click to set main cursor and clear extra cursors
    click_at(codeeditor, qtbot, 6)
    assert not bool(codeeditor.extra_cursors)
    qtbot.keyClick(codeeditor, "b")
    assert codeeditor.toPlainText() == "a01234b5a6789"

    # Don't add another cursor on top of main cursor
    click_at(codeeditor, qtbot, 7, ctrl=True, alt=True)
    assert not bool(codeeditor.extra_cursors)

    # Test removing cursors
    click_at(codeeditor, qtbot, 2, ctrl=True, alt=True)

    # Remove main cursor
    click_at(codeeditor, qtbot, 2, ctrl=True, alt=True)
    assert codeeditor.textCursor().position() == 7


def test_column_add_cursor(codeeditor, qtbot):
    """Test adding a column of extra cursors with ctrl-alt-shift click"""

    codeeditor.set_text("0123456789\n0123456789\n0123456789\n0123456789\n")
    cursor = codeeditor.textCursor()

    # Move main cursor to bottom left
    cursor.movePosition(
        QTextCursor.MoveOperation.Down,
        QTextCursor.MoveMode.MoveAnchor,
        3
    )
    codeeditor.setTextCursor(cursor)

    # Column cursor click at top row 6th column
    click_at(codeeditor, qtbot, 6, ctrl=True, alt=True, shift=True)

    assert codeeditor.extra_cursors
    assert len(codeeditor.all_cursors) == 4
    for cursor in codeeditor.all_cursors:
        assert cursor.selectedText() == "012345"


def test_settings_toggle(codeeditor, qtbot):
    """Test toggling multicursor support in settings"""

    codeeditor.set_text("0123456789\n0123456789\n")
    click_at(codeeditor, qtbot, 6, ctrl=True, alt=True)

    # A cursor was added
    assert bool(codeeditor.extra_cursors)
    codeeditor.toggle_multi_cursor(False)

    # Extra cursors removed on settings toggle
    assert not bool(codeeditor.extra_cursors)
    click_at(codeeditor, qtbot, 3, ctrl=True, alt=True)

    # Extra cursors not added wnen settings "multi-cursor enabled" is False
    assert not bool(codeeditor.extra_cursors)
    click_at(codeeditor, qtbot, 13, ctrl=True, alt=True, shift=True)

    # Column cursors not added wnen settings "multi-cursor enabled" is False
    assert not bool(codeeditor.extra_cursors)


def test_extra_selections_decoration(codeeditor, qtbot):
    """Ensure text decorations are created to paint extra cursor selections."""

    codeeditor.set_text("0123456789\n0123456789\n0123456789\n0123456789\n")
    cursor = codeeditor.textCursor()
    cursor.movePosition(
        QTextCursor.MoveOperation.Down,
        QTextCursor.MoveMode.MoveAnchor,
        3
    )
    codeeditor.setTextCursor(cursor)
    click_at(codeeditor, qtbot, 6, ctrl=True, alt=True, shift=True)
    selections = codeeditor.get_extra_selections("extra_cursor_selections")
    assert len(selections) == 3


def test_multi_cursor_verticalMovementX(codeeditor, qtbot):
    """Ensure cursors keep their column position when moving up and down."""
    codeeditor.set_text("012345678\n012345678\n\n012345678\n012345678\n")
    click_at(codeeditor, qtbot, 4)
    click_at(codeeditor, qtbot, 14, ctrl=True, alt=True)
    for _ in range(3):
        qtbot.keyClick(codeeditor, Qt.Key.Key_Down)
    assert codeeditor.extra_cursors[0].position() == 25
    assert codeeditor.textCursor().position() == 35

    for _ in range(3):
        qtbot.keyClick(codeeditor, Qt.Key.Key_Up)
    assert codeeditor.extra_cursors[0].position() == 4
    assert codeeditor.textCursor().position() == 14


def test_overwrite_mode(codeeditor, qtbot):
    """
    Multi-cursor rendering requires overwrite mode be handled manually as there
    is no way to hide the primary textCursor with overwriteMode, and there is
    no way to sync the blinking of extra cursors with native rendering.

    Test overwrite mode (insert key)
    """

    codeeditor.set_text("0123456789\n0123456789\n")
    click_at(codeeditor, qtbot, 4)
    qtbot.keyClick(codeeditor, Qt.Key.Key_Insert)

    # Overwrite mode is tracked manually, ensure the Qt property is False
    assert not codeeditor.overwriteMode()
    assert codeeditor.overwrite_mode

    # Test overwrite mode functionality for single cursor
    qtbot.keyClick(codeeditor, Qt.Key.Key_A)
    assert codeeditor.toPlainText() == "0123a56789\n0123456789\n"

    # Test overwrite mode for multiple cursors
    click_at(codeeditor, qtbot, 16, ctrl=True, alt=True)
    qtbot.keyClick(codeeditor, Qt.Key.Key_B)
    assert codeeditor.toPlainText() == "0123ab6789\n01234b6789\n"

    # Test returning to insert mode
    qtbot.keyClick(codeeditor, Qt.Key.Key_Insert)
    assert not codeeditor.overwriteMode()
    assert not codeeditor.overwrite_mode
    qtbot.keyClick(codeeditor, Qt.Key.Key_C)
    assert codeeditor.toPlainText() == "0123abc6789\n01234bc6789\n"

# TODO test folded code
#    extra cursor movement (skip past hidden blocks)
#    typing on a folded line (should be read-only)
#    delete & backspace (& delete line shortcut) should delete folded section
#    move/duplicate line up/down should unfold current and previous/next lines


# def test_drag_and_drop(codeeditor, qtbot):
#     # test drag & drop cursor rendering
#     codeeditor.set_text("0123456789\nabcdefghij\n")
#     cursor = codeeditor.textCursor()
#     cursor.movePosition(cursor.MoveOperation.NextBlock,
#                         cursor.MoveMode.KeepAnchor)

#     assert codeeditor._drag_cursor is None
#     point = QPoint(*codeeditor.get_coordinates(5))
#     qtbot.mousePress(codeeditor.viewport(),
#                      Qt.MouseButton.LeftButton,
#                      pos=point)

#     point = QPoint(*codeeditor.get_coordinates(22))
#     # TODO not working: this doesn't generate a DragEnter event or DragMove
#     #    events. Why?
#     qtbot.mouseMove(codeeditor.viewport(),
#                     pos=point)
#     assert codeeditor._drag_cursor is not None
#     qtbot.mouseRelease(codeeditor.viewport(),
#                        Qt.MouseButton.LeftButton,
#                        pos=point)
#     assert codeeditor._drag_cursor is None
#     assert codeeditor.toPlainText() == "abcdefghij\n0123456789\n"

# fmt: off
# Disable formatting so that Black/Ruff don't incorrectly format the multiline
# strings below.
def test_smart_text(codeeditor, qtbot):
    """
    Test smart text features: Smart backspace, whitespace insertion, colon
    insertion, parenthesis and quote matching.
    """

    # Closing paren was inserted?
    codeeditor.set_text("def test1\ndef test2\n")
    click_at(codeeditor, qtbot, 9)
    click_at(codeeditor, qtbot, 19, ctrl=True, alt=True)
    qtbot.keyClick(codeeditor, Qt.Key.Key_ParenLeft)
    assert codeeditor.toPlainText() == ("def test1()\ndef test2()\n")

    # Typing close paren advances cursor without adding extra paren?
    qtbot.keyClick(codeeditor, Qt.Key.Key_ParenRight)
    assert codeeditor.toPlainText() == ("def test1()\ndef test2()\n")

    # Auto colon and indent?
    qtbot.keyClick(codeeditor, Qt.Key.Key_Return)
    assert codeeditor.toPlainText() == (
        "def test1():\n"
        "    \n"
        "def test2():\n"
        "    \n"
    )

    # Add some extraneous whitespace
    qtbot.keyClick(codeeditor, Qt.Key.Key_Tab)
    qtbot.keyClick(codeeditor, Qt.Key.Key_Tab)
    assert codeeditor.toPlainText() == (
        "def test1():\n"
        "            \n"
        "def test2():\n"
        "            \n"
    )

    # Smart backspace to correct indent?
    qtbot.keyClick(codeeditor, Qt.Key.Key_Backspace)
    assert codeeditor.toPlainText() == (
        "def test1():\n"
        "    \n"
        "def test2():\n"
        "    \n"
    )

    for cursor in codeeditor.all_cursors:
        cursor.insertText("return")
    assert codeeditor.toPlainText() == (
        "def test1():\n"
        "    return\n"
        "def test2():\n"
        "    return\n"
    )

    # Automatic quote
    codeeditor.set_close_quotes_enabled(True)
    qtbot.keyClick(codeeditor, Qt.Key.Key_Space)
    qtbot.keyClick(codeeditor, Qt.Key.Key_Apostrophe)
    assert codeeditor.toPlainText() == (
        "def test1():\n"
        "    return ''\n"
        "def test2():\n"
        "    return ''\n"
    )

    # Automatic close quote
    qtbot.keyClick(codeeditor, Qt.Key.Key_Apostrophe)
    assert codeeditor.toPlainText() == (
        "def test1():\n"
        "    return ''\n"
        "def test2():\n"
        "    return ''\n"
    )

    # Automatic dedent?
    qtbot.keyClick(codeeditor, Qt.Key.Key_Return)
    assert codeeditor.toPlainText() == (
        "def test1():\n"
        "    return ''\n"
        "\n"
        "def test2():\n"
        "    return ''\n"
        "\n"
    )

# fmt: on

# ---- shortcuts

# def test_disable_code_completion(completions_codeeditor, qtbot):
#     code_editor, _ = completions_codeeditor
#     code_editor.set_text("def test1():\n"
#                          "    return\n")
#     completion = code_editor.completion_widget
#     delay = 50
#     code_editor.toggle_code_snippets(False)

#     code_editor.moveCursor(QTextCursor.MoveOperation.End)
#     qtbot.keyClicks(code_editor, "tes", delay=delay)
#     qtbot.wait(2000)  # keyboard idle + wait for completion time
#     assert not completion.isHidden()
#     # TODO doesn't hide completions, why?
#     qtbot.keyClick(code_editor, Qt.Key.Key_Escape)
#     # TODO doesn't hide completions, why?
#     click_at(code_editor, qtbot, 0)
#     qtbot.wait(1000)
#     click_at(code_editor, qtbot, 27, ctrl=True, alt=True)
#     qtbot.keyClicks(code_editor, "t", delay=delay)
#     qtbot.wait(1000)  # keyboard idle + wait for completion time
#     # XXX fails because first completions box is never hidden.
#     assert completion.isHidden()
#     # TODO test ctrl-space shortcut too

def test_move_line(codeeditor, qtbot):
    """Test multi-cursor move line up and down shortcut"""

    codeeditor.set_text("\n".join("123456"))
    click_at(codeeditor, qtbot, 4, ctrl=True, alt=True)

    call_shortcut(codeeditor, "move line down")
    call_shortcut(codeeditor, "move line down")
    assert codeeditor.toPlainText() == "\n".join("241536")

    call_shortcut(codeeditor, "move line up")
    assert codeeditor.toPlainText() == "\n".join("214356")


def test_duplicate_line(codeeditor, qtbot):
    """Test multi-cursor duplicate line up and down shortcut"""

    codeeditor.set_text("\n".join("123456"))
    click_at(codeeditor, qtbot, 4, ctrl=True, alt=True)

    call_shortcut(codeeditor, "duplicate line down")
    assert codeeditor.toPlainText() == "\n".join("11233456")
    assert codeeditor.textCursor().position() == 8
    assert codeeditor.extra_cursors[0].position() == 2

    call_shortcut(codeeditor, "duplicate line up")
    assert codeeditor.toPlainText() == "\n".join("1112333456")
    assert codeeditor.textCursor().position() == 10
    assert codeeditor.extra_cursors[0].position() == 2


def test_delete_line(codeeditor, qtbot):
    """Test delete line shortcut"""

    codeeditor.set_text("\n".join("123456"))
    click_at(codeeditor, qtbot, 4, ctrl=True, alt=True)
    call_shortcut(codeeditor, "delete line")
    assert codeeditor.toPlainText() == "\n".join("2456")


def test_goto_new_line(codeeditor, qtbot):
    """Test 'go to new line' shortcut"""

    codeeditor.set_text("\n".join("123456"))
    click_at(codeeditor, qtbot, 4, ctrl=True, alt=True)
    call_shortcut(codeeditor, "go to new line")
    assert codeeditor.toPlainText() == "1\n\n2\n3\n\n4\n5\n6"


# TODO test goto line number / definition / next cell / previous cell
# TODO test toggle comment, blockcomment, unblockcomment
# TODO test transform to UPPER / lower case
# TODO test indent / unindent
# TODO test start/end of document
# TODO test start/end of line
# TODO test prev/next char/word
# TODO test next/prev warning
# TODO test killring
# TODO test undo/redo
# TODO test cut copy paste
# TODO test delete
# TODO test select all
# TODO test docstring
# TODO test autoformatting
# TODO test enter inline array/table
# TODO test inspect current object
# TODO test last edit location
# TODO test next/prev cursor position
# TODO test run Cell (and advance)
# TODO test run selection (and advance)(from line)(in debugger)

if __name__ == '__main__':
    pytest.main(['test_multicursor.py'])



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Tests for the Application plugin.
"""

# Standard library imports
import os.path as osp
from unittest.mock import MagicMock, Mock, patch, ANY

# Third party imports
import pytest

# Local imports
from spyder.api.plugins import Plugins, SpyderDockablePlugin
from spyder.plugins.application.api import ApplicationActions


def test_focused_plugin(application_plugin):
    """
    Test that focused_plugin is initially None and that it is set after
    receiving sig_focused_plugin_changed.
    """
    assert application_plugin.focused_plugin is None
    mock_plugin = Mock()
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin)
    assert application_plugin.focused_plugin == mock_plugin


@pytest.mark.parametrize('can_search', [True, False])
@pytest.mark.parametrize(
    'action_name, editor_function_name, plugin_function_name',
    [
        ('find_action', 'find', 'find'),
        ('find_next_action', 'find_next', 'find_next'),
        ('find_previous_action', 'find_previous', 'find_previous'),
        ('replace_action', 'replace', 'replace'),
    ],
)
def test_search_actions(
    application_plugin, action_name, editor_function_name,
    plugin_function_name, can_search
):
    """
    Test that triggering search actions calls the corresponding function in the
    Editor plugin.
    """
    mock_plugin = Mock(CAN_HANDLE_SEARCH_ACTIONS=can_search)
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin)

    container = application_plugin.get_container()
    action = getattr(container, action_name)
    action.trigger()

    if can_search:
        editor_function = getattr(mock_plugin, plugin_function_name)
        editor_function.assert_called()
    else:
        application_plugin.get_plugin.assert_called_with(Plugins.Editor)
        editor_plugin = application_plugin.get_plugin.return_value
        editor_function = getattr(editor_plugin, editor_function_name)
        editor_function.assert_called()


def test_enable_search_action(application_plugin):
    """
    Test that enable_search_action enables or disabled the specified search
    action on the plugin, and that switching plugins updates whether actions
    are enabled according to previous calls to enable_search_action.
    """
    container = application_plugin.get_container()
    mock_plugin1 = Mock(CAN_HANDLE_EDIT_ACTIONS=True)
    mock_plugin2 = Mock()

    # Initially, actions are enabled
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin1)
    assert container.find_action.isEnabled() is True

    # Disabling the Find Text action in the active plugin works
    application_plugin.enable_search_action(
        ApplicationActions.FindText, False, mock_plugin1.NAME
    )
    assert container.find_action.isEnabled() is False

    # After changing to another plugin, the Find Text action is enabled again
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin2)
    assert container.find_action.isEnabled() is True

    # Disabling the Find Text action in the second plugin (which has focus)
    # works
    application_plugin.enable_search_action(
        ApplicationActions.FindText, False, mock_plugin2.NAME
    )
    assert container.find_action.isEnabled() is False

    # Enabling the Find Text action in the first plugin has no immediate effect
    application_plugin.enable_search_action(
        ApplicationActions.FindText, True, mock_plugin1.NAME
    )
    assert container.find_action.isEnabled() is False

    # When changing to the first plugin, the Find Text action is enabled
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin1)
    assert container.find_action.isEnabled() == True


@pytest.mark.parametrize('can_edit', [True, False])
@pytest.mark.parametrize(
    'action_name, editor_function_name, plugin_function_name',
    [
        ('undo_action', 'undo', 'undo'),
        ('redo_action', 'redo', 'redo'),
        ('cut_action', 'cut', 'cut'),
        ('copy_action', 'copy', 'copy'),
        ('paste_action', 'paste', 'paste'),
        ('select_all_action', 'select_all', 'select_all'),
    ],
)
def test_edit_actions(
    application_plugin, action_name, editor_function_name,
    plugin_function_name, can_edit
):
    """
    Test that triggering edit actions calls the corresponding function in the
    Editor plugin.
    """
    mock_plugin = Mock(CAN_HANDLE_EDIT_ACTIONS=can_edit)
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin)

    container = application_plugin.get_container()
    action = getattr(container, action_name)
    action.trigger()

    if can_edit:
        editor_function = getattr(mock_plugin, plugin_function_name)
        editor_function.assert_called()
    else:
        application_plugin.get_plugin.assert_called_with(Plugins.Editor)
        editor_plugin = application_plugin.get_plugin.return_value
        editor_function = getattr(editor_plugin, editor_function_name)
        editor_function.assert_called()


def test_enable_edit_action(application_plugin):
    """
    Test that enable_edit_action enables or disabled the specified edit action
    on the plugin, and that switching plugins updates whether actions are
    enabled according to previous calls to enable_edit_action.
    """
    container = application_plugin.get_container()
    mock_plugin1 = Mock(CAN_HANDLE_EDIT_ACTIONS=True)
    mock_plugin2 = Mock()

    # Initially, actions are enabled
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin1)
    assert container.undo_action.isEnabled() is True

    # Disabling the Save action in the active plugin works
    application_plugin.enable_edit_action(
        ApplicationActions.Undo, False, mock_plugin1.NAME
    )
    assert container.undo_action.isEnabled() is False

    # After changing to another plugin, the Save action is enabled again
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin2)
    assert container.undo_action.isEnabled() is True

    # Disabling the Save action in the second plugin (which has focus) works
    application_plugin.enable_edit_action(
        ApplicationActions.Undo, False, mock_plugin2.NAME
    )
    assert container.undo_action.isEnabled() is False

    # Enabling the Save action in the first plugin has no immediate effect
    application_plugin.enable_edit_action(
        ApplicationActions.Undo, True, mock_plugin1.NAME
    )
    assert container.undo_action.isEnabled() is False

    # When changing to the first plugin, the Save action is enabled
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin1)
    assert container.undo_action.isEnabled() == True


@pytest.mark.parametrize('can_file', [True, False])
@pytest.mark.parametrize(
    'action_name, editor_function_name, plugin_function_name',
    [
        ('new_action', 'new', 'create_new_file'),
        ('open_last_closed_action', 'open_last_closed', 'open_last_closed_file'),
        ('save_action', 'save', 'save_file'),
        ('save_all_action', 'save_all', 'save_all'),
        ('save_as_action', 'save_as', 'save_file_as'),
        ('save_copy_as_action', 'save_copy_as', 'save_copy_as'),
        ('revert_action', 'revert_file', 'revert_file'),
        ('close_file_action', 'close_file', 'close_file'),
        ('close_all_action', 'close_all_files', 'close_all'),
    ],
)
def test_file_actions(
    application_plugin, action_name, editor_function_name,
    plugin_function_name, can_file
):
    """
    Test that triggering file actions calls the corresponding function in the
    Editor plugin.
    """
    mock_plugin = Mock(CAN_HANDLE_FILE_ACTIONS=can_file)
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin)

    container = application_plugin.get_container()
    action = getattr(container, action_name)
    action.trigger()

    if can_file:
        editor_function = getattr(mock_plugin, plugin_function_name)
        editor_function.assert_called()
    else:
        application_plugin.get_plugin.assert_called_with(Plugins.Editor)
        editor_plugin = application_plugin.get_plugin.return_value
        editor_function = getattr(editor_plugin, editor_function_name)
        editor_function.assert_called()


def test_open_file_action(application_plugin):
    """
    Test that triggering the "Open file" action creates a QFileDialog.
    Assume that there are two plugins, the editor plugin and another plugin
    that says it can handle .xyz files. Check that if the user selects one
    file with no extension and one with an .xyz extension, then the load()
    function in the Editor plugin is called with the first file and the
    open_file() function in the other plugin is called with the second file.
    """
    mock_QFileDialog = Mock()
    mock_QFileDialog.return_value.selectedFiles.return_value = [
        '/home/file1',
        '/home/file2.xyz',
    ]

    xyz_plugin = Mock(spec=SpyderDockablePlugin, FILE_EXTENSIONS=['.xyz'])

    editor_plugin = Mock()
    editor_plugin.get_current_filename.return_value = 'current-file'

    def my_get_plugin(name):
        return {'xyz': xyz_plugin, Plugins.Editor: editor_plugin}[name]

    mock_registry = MagicMock()
    mock_registry.__iter__.side_effect = lambda: iter(['xyz', Plugins.Editor])
    mock_registry.get_plugin.side_effect = my_get_plugin

    application_plugin.get_plugin.side_effect = my_get_plugin

    # Note: container.open_file_using_dialog() behaves differently under pytest
    with patch('spyder.plugins.application.container.QFileDialog', mock_QFileDialog):
        with patch('spyder.plugins.application.plugin.PLUGIN_REGISTRY', mock_registry):
            container = application_plugin.get_container()
            container.open_action.trigger()

    mock_QFileDialog.assert_called()
    editor_plugin.load.assert_called_with(osp.normpath('/home/file1'))
    xyz_plugin.open_file.assert_called_with(osp.normpath('/home/file2.xyz'))


@pytest.mark.parametrize('plugin_filename', ['plugin', None])
def test_open_file_using_dialog(application_plugin, plugin_filename):
    """
    That that open_file_using_dialog asks the currently focused plugin for the
    current file name, and that if the result is None the Editor plugin is
    asked as a fallback.
    """
    mock_plugin = Mock()
    mock_plugin.get_current_filename.return_value = plugin_filename

    mock_editor = Mock()
    mock_editor.get_current_filename.return_value = 'editor'

    application_plugin.focused_plugin = mock_plugin
    application_plugin.get_plugin.return_value = mock_editor

    container = application_plugin.get_container()
    with patch.object(container, 'open_file_using_dialog') as mock:
        application_plugin.open_file_using_dialog()

    if plugin_filename:
        mock.assert_called_with(plugin_filename, ANY)
    else:
        application_plugin.get_plugin.assert_called_with(Plugins.Editor)
        mock.assert_called_with('editor', ANY)


def test_enable_file_action(application_plugin):
    """
    Test that enable_file_action enables or disabled the specified file action
    on the plugin, and that switching plugins updates whether actions are
    enabled according to previous calls to enable_file_action.
    """
    container = application_plugin.get_container()
    mock_plugin1 = Mock(CAN_HANDLE_FILE_ACTIONS=True)
    mock_plugin2 = Mock()

    # Initially, actions are enabled
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin1)
    assert container.save_action.isEnabled() is True

    # Disabling the Save action in the active plugin works
    application_plugin.enable_file_action(
        ApplicationActions.SaveFile, False, mock_plugin1.NAME
    )
    assert container.save_action.isEnabled() is False

    # After changing to another plugin, the Save action is enabled again
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin2)
    assert container.save_action.isEnabled() is True

    # Disabling the Save action in the second plugin (which has focus) works
    application_plugin.enable_file_action(
        ApplicationActions.SaveFile, False, mock_plugin2.NAME
    )
    assert container.save_action.isEnabled() is False

    # Enabling the Save action in the first plugin has no immediate effect
    application_plugin.enable_file_action(
        ApplicationActions.SaveFile, True, mock_plugin1.NAME
    )
    assert container.save_action.isEnabled() is False

    # When changing to the first plugin, the Save action is enabled
    application_plugin.sig_focused_plugin_changed.emit(mock_plugin1)
    assert container.save_action.isEnabled() == True



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

import os
from functools import lru_cache
import logging
from packaging.version import parse

import pytest

from spyder.config.base import running_in_ci
from spyder.plugins.updatemanager import workers
from spyder.plugins.updatemanager.workers import (
    UpdateType, get_asset_info, WorkerUpdate
)
from spyder.plugins.updatemanager.widgets import update
from spyder.plugins.updatemanager.widgets.update import UpdateManagerWidget

logging.basicConfig()

__get_github_releases = workers.get_github_releases


@lru_cache()
def _get_github_releases():
    return __get_github_releases(
        ("v6.0.0b3", "v6.0.0rc1", "v6.0.7", "v6.1.0a3")
    )


workers.get_github_releases = _get_github_releases
workers.get_github_releases()  # Run once to cache result for tests


@pytest.fixture(autouse=True)
def capture_logging(caplog):
    # Capture >=DEBUG logging messages for spyder.plugins.updatemanager.
    # Messages will be reported at the end of the pytest run for failed tests.
    caplog.set_level(10, "spyder.plugins.updatemanager")


# ---- Test WorkerUpdate

@pytest.mark.parametrize("version", ["1.0.0", "1000.0.0"])
@pytest.mark.parametrize(
    "channel", [
        ("pkgs/main", "https://repo.anaconda.com/pkgs/main"),
        ("conda-forge", "https://conda.anaconda.org/conda-forge"),
    ]
)
def test_updates(qtbot, mocker, caplog, version, channel):
    """
    Test whether or not we offer updates according to the current Spyder
    version and package installation channel.

    Uses UpdateManagerWidget in order to also test QThread.
    """
    mocker.patch.object(update, "__version__", new=version)
    # Do not execute start_update after check_update completes.
    mocker.patch.object(
        UpdateManagerWidget, "start_update", new=lambda x: None
    )
    mocker.patch.object(workers, "CURRENT_VERSION", new=parse(version))
    mocker.patch.object(
        workers, "get_spyder_conda_channel", return_value=channel
    )

    um = UpdateManagerWidget(None)
    um.start_check_update()
    qtbot.waitUntil(um.update_thread.isFinished, timeout=10000)

    if version.split('.')[0] == '1':
        assert um.update_worker.asset_info is not None
    else:
        assert um.update_worker.asset_info is None


@pytest.mark.parametrize("version", ["5.5.6", "6.0.0a1"])
@pytest.mark.parametrize(
    "stable_only,release,expected",
    [
        (True, "6.0.7", "6.0.7"),
        (True, "6.1.0a3", "6.0.7"),
        (True, "6.0.0rc1", None),
        (False, "6.0.7", "6.0.7"),
        (False, "6.0.0rc1", "6.0.0rc1")
    ]
)
def test_update_non_stable(
    qtbot, mocker, version, stable_only, release, expected
):
    """Test we offer unstable updates."""
    mocker.patch.object(workers, "CURRENT_VERSION", new=parse(version))

    worker = WorkerUpdate(stable_only)
    release = parse(release)

    worker._check_update_available(release)

    if expected is not None:
        expected = parse(expected)
        assert worker.asset_info["version"] == expected
    else:
        assert worker.asset_info is None


def test_update_no_asset(qtbot, mocker):
    """Test update availability when asset is not available"""
    mocker.patch.object(workers, "CURRENT_VERSION", new=parse("6.0.0b2"))
    mocker.patch.object(workers, "is_conda_based_app", return_value=True)

    worker = WorkerUpdate(False)

    # 6.0.0b3 does not have correct minor/micro update asset
    worker._check_update_available(parse("6.0.0b3"))
    assert worker.asset_info is None

    # 6.0.0rc1 is first release to have correct minor/micro update asset
    worker._check_update_available(parse("6.0.0rc1"))
    assert worker.asset_info is not None
    assert worker.asset_info["version"] == parse("6.0.0rc1")


@pytest.mark.parametrize(
    "app,version,release,update_type",
    [
        (True, "6.0.0", "6.0.7", UpdateType.Micro),
        (True, "6.0.0b3", "6.0.0rc1", UpdateType.Minor),
        (True, "6.0.0", "6.1.0a3", UpdateType.Minor),
        (True, "5.0.0", "6.0.7", UpdateType.Major),
        (False, "6.0.0", "6.0.7", UpdateType.Major),
        (False, "6.0.0", "6.1.0a3", UpdateType.Major),
        (False, "5.0.0", "6.0.7", UpdateType.Major)
    ]
)
def test_get_asset_info(qtbot, mocker, app, version, release, update_type):
    mocker.patch.object(workers, "CURRENT_VERSION", new=parse(version))
    mocker.patch.object(workers, "is_conda_based_app", return_value=app)

    worker = WorkerUpdate(False)
    worker._check_update_available(parse(release))
    info = worker.asset_info

    assert info['update_type'] == update_type

    if update_type == "major" or not app:
        assert info['url'].endswith(('.exe', '.pkg', '.sh'))
        assert info['filename'].endswith(('.exe', '.pkg', '.sh'))
    else:
        assert info['url'].endswith(".zip")
        assert info['filename'].endswith(".zip")


# ---- Test WorkerDownloadInstaller

@pytest.mark.skip(reason="Re-enable when alternate repo is available")
@pytest.mark.skipif(not running_in_ci(), reason="Download only in CI")
@pytest.mark.parametrize("version", ["5.5.6", "6.0.0"])
def test_download(qtbot, mocker, version):
    """
    Test download spyder installer.

    Uses UpdateManagerWidget in order to also test QThread.
    """
    version = parse(version)
    mocker.patch.object(workers, "CURRENT_VERSION", new=version)

    releases = workers.get_github_releases()
    release_info = releases[parse("6.1.0a1")]

    um = UpdateManagerWidget(None)
    um.asset_info = get_asset_info(release_info, version, False)
    um._set_installer_path()

    # Do not execute _start_install after download completes.
    mocker.patch.object(
        UpdateManagerWidget, "_confirm_install", new=lambda x: None
    )

    um._start_download()
    qtbot.waitUntil(um.download_thread.isFinished, timeout=60000)

    assert os.path.exists(um.installer_path)


if __name__ == "__main__":
    pytest.main()



# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""IPython console and Spyder Remote Client integration tests."""

# Third party imports
from flaky import flaky
import pytest

# Local imports
from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.plugins.remoteclient.tests.conftest import (
    await_future,
    mark_remote_test,
)


@AsyncDispatcher(loop="test")
async def list_kernels(remote_client_plugin, server_id):
    async with remote_client_plugin.get_jupyter_api(server_id) as jupyter_api:
        return await jupyter_api.list_kernels()


@mark_remote_test
class TestIpythonConsole:
    def test_start_stop_kernel(
        self, ipyconsole, remote_client, remote_client_id, qtbot
    ):
        """Starts and stops a kernel on the remote server."""
        ipyconsole.get_widget().create_ipyclient_for_server(remote_client_id)
        shell = ipyconsole.get_current_shellwidget()

        qtbot.waitUntil(
            lambda: shell.spyder_kernel_ready
            and shell._prompt_html is not None,
            timeout=180000,  # longer timeout for installation
        )

        ipyconsole.get_widget().close_client()

        assert await_future(
            list_kernels(remote_client, remote_client_id),
            timeout=5
        ) == []

    def test_restart_kernel(self, remote_shell, ipyconsole, qtbot):
        """Test that kernel is restarted correctly."""
        # Do an assignment to verify that it's not there after restarting
        with qtbot.waitSignal(remote_shell.executed):
            remote_shell.execute("a = 10")

        with qtbot.waitSignal(
            remote_shell.sig_kernel_is_ready,
            timeout=4000
        ):
            ipyconsole.get_widget().restart_action.trigger()

        assert not remote_shell.is_defined("a")

        with qtbot.waitSignal(remote_shell.executed):
            remote_shell.execute("b = 10")

        # Make sure that kernel is responsive after restart
        qtbot.waitUntil(lambda: remote_shell.is_defined("b"), timeout=2000)

    def test_interrupt_kernel(self, remote_shell, qtbot):
        """Test that the kernel correctly interrupts."""
        loop_string = "while True: pass"

        with qtbot.waitSignal(remote_shell.executed):
            remote_shell.execute("b = 10")

        remote_shell.execute(loop_string)

        qtbot.wait(500)

        remote_shell.interrupt_kernel()

        qtbot.wait(1000)

        # Make sure that kernel didn't die
        assert remote_shell.get_value("b") == 10

    def test_kernel_kill(self, remote_shell, qtbot):
        """Test that the kernel correctly restarts after a kill."""
        crash_string = (
            "import os, signal; os.kill(os.getpid(), signal.SIGTERM)"
        )

        with qtbot.waitSignal(remote_shell.sig_prompt_ready, timeout=30000):
            remote_shell.execute(crash_string)

        assert "The kernel died, restarting..." in remote_shell._control.toPlainText()

        with qtbot.waitSignal(remote_shell.executed):
            remote_shell.execute("b = 10")

        # Make sure that kernel is responsive after restart
        qtbot.waitUntil(lambda: remote_shell.is_defined("b"), timeout=2000)


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tests for menus.py .
"""

from spyder.api.widgets.menus import MENU_SEPARATOR, SpyderAction, SpyderMenu


def test_add_action_with_before_section(qtbot):
    """
    Check that the actions in a menu are in the right order if
    before_section is used.
    """
    action1 = SpyderAction('action1', action_id='action1')
    action2 = SpyderAction('action2', action_id='action2')
    action3 = SpyderAction('action3', action_id='action3')
    action4 = SpyderAction('action4', action_id='action4')

    menu = SpyderMenu()
    menu.add_action(action1, section='section1', before_section='section2')
    menu.add_action(action4, section='section4')
    menu.add_action(action2, section='section2', before_section='section3')
    menu.add_action(action3, section='section3', before_section='section4')

    result = menu.get_actions()
    expected = [
        action1,
        MENU_SEPARATOR,
        action2,
        MENU_SEPARATOR,
        action3,
        MENU_SEPARATOR,
        action4,
        MENU_SEPARATOR
    ]
    assert result == expected



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

from argparse import ArgumentDefaultsHelpFormatter, RawDescriptionHelpFormatter
from logging import Formatter, StreamHandler, getLogger
from pathlib import Path

fmt = Formatter('%(asctime)s [%(levelname)s] [%(name)s] -> %(message)s')
h = StreamHandler()
h.setFormatter(fmt)
logger = getLogger('Installer')
logger.addHandler(h)
logger.setLevel('INFO')

HERE = Path(__file__).parent
SPYREPO = HERE.parent
RESOURCES = HERE / "resources"
BUILD = HERE / "build"
DIST = HERE / "dist"


class DocFormatter(
    RawDescriptionHelpFormatter,
    ArgumentDefaultsHelpFormatter
):
    pass



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Tests for ApplicationContainer.
"""

# Standard library imports
from unittest.mock import patch


def test_add_recent_file(application_plugin):
    """
    Test that add_recent_file adds the given file to the front of recent_files
    and ensures that the list has no duplicates.
    """
    container = application_plugin.get_container()
    assert container.recent_files == []

    container.add_recent_file('file1')
    assert container.recent_files == ['file1']

    container.add_recent_file('file2')
    assert container.recent_files == ['file2', 'file1']

    container.add_recent_file('file1')
    assert container.recent_files == ['file1', 'file2']


def test_clear_recent_files(application_plugin):
    """
    Test that clear_recent_files clears the list in recent_files.
    """
    container = application_plugin.get_container()
    container.add_recent_file('file1')
    assert container.recent_files == ['file1']

    container.clear_recent_files()
    assert container.recent_files == []


def test_update_recent_files_menu(application_plugin):
    """
    Test that update_recent_files_menu() puts the files in recent_files in
    the recent_files_menu (skipping those that do not exist), followed by a
    separator and then other items.
    """

    def mock_isfile(filename):
        return filename in ['file1', 'file2']

    container = application_plugin.get_container()
    container.add_recent_file('file1')
    container.add_recent_file('file2')
    container.add_recent_file('non-existing-file')
    with patch('spyder.plugins.application.container.osp.isfile', mock_isfile):
        container.update_recent_files_menu()

    menu = container.recent_files_menu
    menuitems = [action.text() for action in menu.actions()]
    expected = [
        'file2',
        'file1',
        '',
        'Maximum number of recent files...',
        'Clear this list',
        '',
    ]
    assert menuitems == expected


def test_change_max_recent_files(application_plugin):
    """
    Test that change_max_recent_files changes the value of max_recent_files
    in the config to the number given by the user.
    """
    container = application_plugin.get_container()
    old_value = container.get_conf('max_recent_files')
    with patch(
        'spyder.plugins.application.container.QInputDialog.getInt',
        return_value=(old_value + 1, True),
    ):
        container.change_max_recent_files()

    assert container.get_conf('max_recent_files') == old_value + 1

    container.set_conf('max_recent_files', old_value)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Standard library imports
import ast
import os
import re
import sys
from argparse import ArgumentParser
from configparser import ConfigParser
from datetime import timedelta
from subprocess import check_call
from textwrap import dedent
from time import time

# Third-party imports
from packaging.version import parse, InvalidVersion
from git import Repo, rmtree
from ruamel.yaml import YAML
from setuptools_scm import get_version

# Local imports
from utils import logger as logger, HERE, BUILD, DocFormatter

EXTDEPS = HERE.parent / "external-deps"
REQUIREMENTS = HERE.parent / "requirements"
REQ_MAIN = REQUIREMENTS / 'main.yml'
REQ_WINDOWS = REQUIREMENTS / 'windows.yml'
REQ_MAC = REQUIREMENTS / 'macos.yml'
REQ_LINUX = REQUIREMENTS / 'linux.yml'

SPYPATCHFILE = BUILD / "installers-conda.patch"
PARENT_BRANCH = os.getenv("MATRIX_BRANCH", os.getenv("GITHUB_BASE_REF"))


yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)


def get_spy_feedstock_branch():
    """
    Get the right Spyder/Spyder-kernels feedstock branch according to the
    published Spyder repo git tags.
    """
    for line in (
        (HERE.parent / "spyder" / "__init__.py").read_text().split("\n")
    ):
        if "version_info = " in line:
            version_info = ast.literal_eval(line.split("=")[-1])

    spyder_repo = Repo(HERE.parent)

    # Check if the first two version_info elements are part of a stable release
    # version tag
    feedstock_branch = None
    for tag in spyder_repo.tags:
        if not tag.name.startswith("v"):
            continue

        try:
            pre_release = parse(tag.name[0:]).is_prerelease
        except InvalidVersion:
            pre_release = True

        if (
            not pre_release
            and ".".join(map(str, version_info[:2])) in tag.name
        ):
            feedstock_branch = "main"
            break

    # If the above fails, check the fourth version_info element (i.e. "a2" or
    # "b3")
    if feedstock_branch is None:
        if re.match("(b|rc)[0-9].*", version_info[3]):
            feedstock_branch = "rc"
        else:
            feedstock_branch = "dev"

    return feedstock_branch


class BuildCondaPkg:
    """Base class for building a conda package for conda-based installer"""
    name = None
    norm = True
    source = None
    feedstock = None
    feedstock_branch = None

    def __init__(self, data=None, debug=False):
        data = {} if data is None else data
        self.logger = logger.getChild(self.__class__.__name__)

        self.debug = debug

        self._bld_src = BUILD / self.name
        self._fdstk_path = BUILD / self.feedstock.split("/")[-1]
        self._patchfile = self._fdstk_path / "recipe" / "version.patch"

        self._get_source()
        self._get_version()

        self.data = {'version': self.version}
        self.data.update(data)

        self.recipe_append = {}
        self.recipe_clobber = {}

        self._recipe_patched = False

    def _get_source(self):
        """Clone source and feedstock for building"""
        BUILD.mkdir(exist_ok=True)
        self._cleanup_build()

        if self.source == HERE.parent:
            self._bld_src = self.source
            self.repo = Repo(self.source)
        else:
            # Determine source and commit
            if self.source is not None:
                remote = self.source
                commit = 'HEAD'
            else:
                cfg = ConfigParser()
                cfg.read(EXTDEPS / self.name / '.gitrepo')
                remote = cfg['subrepo']['remote']
                commit = cfg['subrepo']['commit']

            # Clone from source
            kwargs = dict(to_path=self._bld_src)
            self.logger.info("Cloning source...")
            self.repo = Repo.clone_from(remote, **kwargs)
            self.repo.git.checkout(commit)

        # Clone feedstock
        self.logger.info("Cloning feedstock...")
        kwargs = dict(to_path=self._fdstk_path)
        if self.feedstock_branch:
            kwargs.update(branch=self.feedstock_branch)
        feedstock_repo = Repo.clone_from(self.feedstock, **kwargs)
        self.logger.info(
            f"Feedstock branch: {feedstock_repo.active_branch.name}"
        )

    def _cleanup_build(self, debug=False):
        """Remove cloned source and feedstock repositories"""
        if debug:
            self.logger.info("Keeping cloned source and feedstock")
            return

        for src in [self._bld_src, self._fdstk_path]:
            if src.exists() and src != HERE.parent:
                self.logger.info(f"Removing {src}...")
                rmtree(src)

    def _get_version(self):
        """Get source version using setuptools_scm"""
        v = get_version(self._bld_src, normalize=self.norm)
        self.version = v.lstrip('v').split('+')[0]

    def _patch_source(self):
        pass

    def _patch_conda_build_config(self):
        file = self._fdstk_path / "recipe" / "conda_build_config.yaml"
        if not file.exists():
            return

        contents = yaml.load(file.read_text())
        file.rename(file.parent / ("_" + file.name))  # copy of original

        pyver = sys.version_info
        contents['python'] = [f"{pyver.major}.{pyver.minor}.* *_cpython"]
        yaml.dump(contents, file)

        self.logger.info(
            f"Patched 'conda_build_config.yaml' contents:\n{file.read_text()}"
        )

    def _patch_meta(self):
        file = self._fdstk_path / "recipe" / "meta.yaml"
        meta = file.read_text()

        # Replace jinja variable values
        for k, v in self.data.items():
            meta = re.sub(f".*set {k} =.*", f'{{% set {k} = "{v}" %}}', meta)

        # Remove temporary patches
        meta = re.sub(r"^\s*- temp-.+\.patch\n", "", meta, flags=re.MULTILINE)

        file.rename(file.parent / ("_" + file.name))  # keep copy of original
        file.write_text(meta)

        self.logger.info(f"Patched 'meta.yaml' contents:\n{file.read_text()}")

    def _add_recipe_append(self):
        if self._patchfile.exists():
            self.recipe_append.update(
                {"source": {"patches": [self._patchfile.name]}}
            )

        if self.recipe_append:
            file = self._fdstk_path / "recipe" / "recipe_append.yaml"
            yaml.dump(self.recipe_append, file)
            self.logger.info(
                f"'recipe_append.yaml' contents:\n{file.read_text()}"
            )
        else:
            self.logger.info("Skipping 'recipe_append.yaml'.")

    def _add_recipe_clobber(self):
        self.recipe_clobber.update({
            "source": {
                "url": None,
                "sha256": None,
                "path": self._bld_src.as_posix()},
        })

        if self.recipe_clobber:
            file = self._fdstk_path / "recipe" / "recipe_clobber.yaml"
            yaml.dump(self.recipe_clobber, file)
            self.logger.info(
                f"'recipe_clobber.yaml' contents:\n{file.read_text()}"
            )
        else:
            self.logger.info("Skipping 'recipe_clobber.yaml'.")

    def patch_recipe(self):
        """
        Patch conda build recipe

        1. Patch conda_build_config.yaml
        2. Patch meta.yaml
        3. Add recipe_append.yaml
        4. Add recipe_clobber.yaml
        """
        if self._recipe_patched:
            return

        self._patch_conda_build_config()
        self._patch_meta()
        self._add_recipe_append()
        self._add_recipe_clobber()

        self._recipe_patched = True

    def build(self):
        """
        Build the conda package.

        1. Patch source
        2. Patch the recipe
        3. Build the package
        4. Remove cloned repositories
        """
        t0 = time()
        try:
            self._patch_source()
            self.patch_recipe()

            self.logger.info("Building conda package "
                             f"{self.name}={self.version}...")
            check_call([
                "conda", "build",
                "--skip-existing", "--build-id-pat={n}",
                "--no-test", "--no-anaconda-upload",
                str(self._fdstk_path / "recipe")
            ])
        finally:
            self._recipe_patched = False
            self._cleanup_build(self.debug)

            elapse = timedelta(seconds=int(time() - t0))
            self.logger.info(f"Build time = {elapse}")


class SpyderCondaPkg(BuildCondaPkg):
    name = "spyder"
    norm = False
    source = os.environ.get('SPYDER_SOURCE', HERE.parent)
    feedstock = "https://github.com/conda-forge/spyder-feedstock"
    feedstock_branch = get_spy_feedstock_branch()

    def _patch_source(self):
        self.logger.info("Patching Spyder source...")

        file = self._bld_src / "spyder/__init__.py"
        file_text = file.read_text()
        ver_str = tuple(self.version.split('.'))
        file_text = re.sub(
            r'^(version_info = ).*',
            rf'\g<1>{ver_str}',
            file_text,
            flags=re.MULTILINE
        )
        file.write_text(file_text)

        # Only write patch if necessary
        if self.repo.git.diff():
            self.logger.info(f"Creating {self._patchfile.name}...")
            self.repo.git.diff(output=self._patchfile.as_posix())
            self.repo.git.stash()

    def patch_recipe(self):
        # Get current Spyder requirements
        spyder_base_reqs = ['python']
        spyder_base_reqs += yaml.load(
            REQ_MAIN.read_text())['dependencies']
        if os.name == 'nt':
            win_requirements = yaml.load(
                REQ_WINDOWS.read_text())['dependencies']
            spyder_base_reqs += win_requirements
            spyder_base_reqs.append('ptyprocess >=0.5')
        elif sys.platform == 'darwin':
            mac_requirements = yaml.load(
                REQ_MAC.read_text())['dependencies']
            if 'python.app' in mac_requirements:
                mac_requirements.remove('python.app')
            spyder_base_reqs += mac_requirements
            spyder_base_reqs.append('__osx')
        else:
            linux_requirements = yaml.load(
                REQ_LINUX.read_text())['dependencies']
            spyder_base_reqs += linux_requirements
            spyder_base_reqs.append('__linux')

        spyder_reqs = [f"spyder-base =={self.version}"]
        for req in spyder_base_reqs.copy():
            if req.startswith(
                ('pyqt ', 'pyqtwebengine ', 'qtconsole ', 'fcitx-qt5 ')
            ):
                spyder_reqs.append(req)
                spyder_base_reqs.remove(req)

            if req.startswith('qtconsole '):
                spyder_base_reqs.append(
                    req.replace('qtconsole', 'qtconsole-base')
                )

        if sys.platform == "darwin":
            spyder_base_reqs.append("__osx")
        if sys.platform.startswith("linux"):
            spyder_base_reqs.append("__linux")
            spyder_reqs.append("__linux")

        self.recipe_clobber.update({
            "requirements": {"run": spyder_base_reqs},
            # Since outputs is a list, the entire list must be reproduced with
            # the current run requirements
            "outputs": [
                {
                    "name": "spyder-base"
                },
                {
                    "name": "spyder",
                    "build": {"noarch": "python"},
                    "requirements": {"run": spyder_reqs},
                    "test": {
                        "requires": ["pip"],
                        "commands": ["spyder -h", "python -m pip check"],
                        "imports": ["spyder"]
                    }
                }
            ]
        })

        super().patch_recipe()


class PylspCondaPkg(BuildCondaPkg):
    name = "python-lsp-server"
    source = os.environ.get('PYTHON_LSP_SERVER_SOURCE')
    feedstock = "https://github.com/conda-forge/python-lsp-server-feedstock"
    feedstock_branch = "main"


class QtconsoleCondaPkg(BuildCondaPkg):
    name = "qtconsole"
    source = os.environ.get('QTCONSOLE_SOURCE')
    feedstock = "https://github.com/conda-forge/qtconsole-feedstock"
    feedstock_branch = "main"


class SpyderKernelsCondaPkg(BuildCondaPkg):
    name = "spyder-kernels"
    source = os.environ.get('SPYDER_KERNELS_SOURCE')
    feedstock = "https://github.com/conda-forge/spyder-kernels-feedstock"
    feedstock_branch = get_spy_feedstock_branch()


PKGS = {
    SpyderCondaPkg.name: SpyderCondaPkg,
    PylspCondaPkg.name: PylspCondaPkg,
    QtconsoleCondaPkg.name: QtconsoleCondaPkg,
    SpyderKernelsCondaPkg.name: SpyderKernelsCondaPkg
}

if __name__ == "__main__":
    parser = ArgumentParser(
        description="Build conda packages to local channel.",
        epilog=dedent(
            """
            This module builds conda packages for Spyder and external-deps for
            inclusion in the conda-based installer. The following classes are
            provided for each package:
                SpyderCondaPkg
                PylspCondaPkg
                QtconsoleCondaPkg
                SpyderKernelsCondaPkg

            Spyder will be packaged from this repository (in its checked-out
            state). qtconsole, spyder-kernels, and python-lsp-server will be
            packaged from the remote and commit specified in their respective
            .gitrepo files in external-deps.

            Alternatively, any external-deps may be packaged from an arbitrary
            git repository (in its checked out state) by setting the
            appropriate environment variable from the following:
                SPYDER_SOURCE
                PYTHON_LSP_SERVER_SOURCE
                QTCONSOLE_SOURCE
                SPYDER_KERNELS_SOURCE
            """
        ),
        formatter_class=DocFormatter
    )
    parser.add_argument(
        '--build', nargs="+", default=list(PKGS.keys()),
        help="Space-separated list of packages to build."
    )
    parser.add_argument(
        '--debug', action='store_true', default=False,
        help="Do not remove cloned sources and feedstocks"
    )

    args = parser.parse_args()

    logger.info(f"Building local conda packages {list(args.build)}...")
    t0 = time()

    for k in args.build:
        pkg = PKGS[k](debug=args.debug)
        pkg.build()

    elapse = timedelta(seconds=int(time() - t0))
    logger.info(f"Total build time = {elapse}")
